defmodule Dp.Reps do

  @moduledoc """
  A ++ operátor hatékonyságának vizsgálata

  @author "hanak@emt.bme.hu"
  @date   "$LastChangedDate: 2021-09-23 07:45:58 +0200 (cs, 23 szept 2021) $$"
  """
  
  # @spec reps1([number]) :: {}
  @spec reps1([number]) :: [number]
  def reps1([]),    do: []
  def reps1([e|es]) do     # A részeredmények a veremben gyűlnek...
    (List.duplicate e, (floor e)) ++ reps1 es
  end

  # def reps2(es), do: List.flatten(reps21 es)
  # def reps21([]),     do: []
  # def reps21([e|es]), do: [(List.duplicate e, (floor e)) | reps21 es]

  def reps2(es), do: reps21 es, []

  defp reps21([],     as), do: Enum.reverse as
  defp reps21([e|es], as), do: reps21 es, List.duplicate(e, floor e) ++ as

  def reps2_1(es), do: reps2_11 (Enum.reverse es), []

  defp reps2_11([],     as), do: as
  defp reps2_11([e|es], as), do: reps2_11 es, List.duplicate(e, floor e) ++ as
  
  def reps3(es), do: reps30 (Enum.reverse es), []
  
  defp reps30([],     as), do: as
  defp reps30([r|rs], as), do: reps30 rs, (reps31 (floor r), r, as)

  defp reps31(0, _r, as), do: as
  defp reps31(i,  r, as),  do: reps31 i-1, r, [r|as]

  # Reps.reps3 (Enum.to_list 1..900)
  # Reps.reps3 (List.duplicate 1,100_000)
  # Reps.reps3 (List.duplicate 100.0,100_000)

  # length(Reps.reps1 (Enum.to_list 799_800..799_900))
  # length(Reps.reps1 (for i <- (Enum.to_list 799_800..799_900), do: i/1))

  # https://www.wyeworks.com/blog/2019/03/01/to-use-or-not-to-use-the-++-operator-in-elixir/

  def timer(f, p) do
    :erlang.garbage_collect()  # collect garbage now not during measurement
    {t, _} = :timer.tc(f, [p]) # get measured runtime, ignore result
    "#{round t/1000} ms"       # return runtime in millisecs
  end

  def stopper(f, p) do
    st = &:erlang.statistics/1 # define alias
    :erlang.garbage_collect()  # collect garbage now not during measurement
    # :timer.sleep(1000)         # wait for a bit (in microsecs)
    st.(:runtime)              # clear runtime since last call
    f.(p)                      # evaluate function
    {_,t} = st.(:runtime)      # ignore total runtime, get runtime since last call
    "#{t} ms"                  # return runtime in millisecs
    # IO.puts("#{t} ms")         # show runtime in millisecs
  end
  
  # Reps.stopper &Reps.reps3/1, (List.duplicate 1, 100_000)
  # Reps.stopper &Reps.reps3/1, (List.duplicate 100.0, 100_000)

  # Reps.timer   &Reps.reps1/1, (Enum.to_list 119_800..119_900)
  # Reps.stopper &Reps.reps1/1, (Enum.to_list 119_800..119_900)
  
  def number_dups_list(numbers) do
    Enum.reduce(
      numbers,
      [],
      fn n, acc ->
	acc ++ List.duplicate(n, n)
      end
    )
  end
  # Reps.number_dups_list(List.duplicate(1,100_000))

  def rev1([]), do: []
  def rev1([x|xs]), do: rev1(xs) ++ [x]

  def rev2(xs), do: rev20 xs, []
  
  defp rev20([], as),     do: as
  defp rev20([x|xs], as), do: rev20 xs, [x|as]
  
  def run_benchmarks do
    inputs = %{
      "Long list with low values" => List.duplicate(1,100_000),
      "Short list with high values" => [1_000_000, 2_000_000]
    }
  
    Benchee.run %{
      "Using ++" =>
    fn(list) -> reps1(list) end,
      "Using ++ and reverse"  =>
      fn(list) -> reps2_1(list) end,
    }, time: 20, warmup: 5, inputs: inputs
  end
  
end
