defmodule Dp.Stack do

  @moduledoc """
  A ++ operátor hatékonyságának vizsgálata

  @author "hanak@emt.bme.hu"
  @date   "$LastChangedDate: 2021-09-23 07:45:58 +0200 (cs, 23 szept 2021) $$"
  """

@type stack :: :empty | {any, stack}

@spec empty() :: s::stack
# s az üres verem
def empty(), do: :empty

@spec is_empty(s::stack) :: b::boolean
# b igaz, ha s üres
def is_empty(:empty), do: true
def is_empty({_,_}),  do: false

@spec push(s::stack, x::any) :: s_new::stack
# Az s verem a tetejére rakott x-szel az s_new verem
def push(:empty, x),       do: {x, :empty}
def push({_x, _s} = s, x), do: {x, s}


@spec pop(s::stack) :: s_old::(stack | nil)
# Ha az s verem nem üres, a tetején lévő elem eltávolításával
# jön létre s_old; ha üres, akkor s_old === nil
def pop(:empty),  do: nil
def pop({_x, s}), do: s

@spec top(s::stack) :: x::(any | nil)
# Ha az s verem nem üres, a tetején lévő elem az x; ha üres,
# akkor x === nil
def top(:empty),  do: nil
def top({x, _s}), do: x

@spec to_list(s::stack) :: ls::list
# ls az s verem elemeit tartalmazó lista LIFO sorrendben
def to_list(:empty), do: []
def to_list({x, s}), do: [ x | to_list(s) ]

def test(1) do
#    push(push(push(empty, 3), 2), 1);
    s1 = push(:empty, 1)
    s2 = push(s1, 2)
    s3 = push(s2, 3)
    s3
end
    
def test(2_1) do
    s = push(push(push(:empty, 1), 2), 3)
    :io.format("stack: ~w, top: ~w, pop: ~w~n", [s, top(s), pop(s)])
end
    
def test(2_2) do
  # wr = fn(s) -> :io.format("stack: ~w, top: ~w, pop: ~w~n", [s, top(s), pop(s)]) end
  wr = &(:io_lib.format("stack: ~w, top: ~w, pop: ~w~n", [&1,top(&1),pop(&1)]) |> to_string)
  :empty
  |> push(1)
  |> push(2)
  |> push(3)
  |> wr.()
end

# karakterláncot megfordít
def test(3_1) do
    stack = List.foldl('szoveg', :empty, fn(c, a) -> push(a, c) end)
    to_list(stack)
end

# karakterláncot megfordít
def test(3_2) do
  'szoveg'
  |> List.foldl(:empty, &(push(&2, &1)))
  |>  to_list
end

# @spec print(Stack) :: none
# Kiírja a kimenetre a Stack tartalmát
# print(empty) ->
#     io:format("empty~n");
# print({X,S}) ->
#     io:format("~p,", [X]),
#     print(S)

end
